import logging
from concurrent.futures import ThreadPoolExecutor
from pydantic import BaseSettings

from .toolkit.testing import TestDaoFactory, TestData
from .server import get_api, serve, ApiSettings
from prometheus_client import make_asgi_app, multiprocess, CollectorRegistry


class Settings(BaseSettings):
    api: ApiSettings = ApiSettings()

    evaluation_pool_size: int = 5


settings = Settings()


def get_dao_factory():
    return TestDaoFactory(TestData())


def get_dao():
    try:
        dao = get_dao_factory().get_dao()
        yield dao
    finally:
        dao.close()


def get_executor_pool():
    try:
        evaluation_pool_size = settings.evaluation_pool_size
        evaluation_pool = ThreadPoolExecutor(max_workers=evaluation_pool_size)
        yield evaluation_pool
    finally:
        pass


def make_metrics_app():
    if settings.api.web_workers == 1:
        return make_asgi_app()
    registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(registry)
    return make_asgi_app(registry=registry)


api = get_api(settings.api, get_dao, get_executor_pool)
metrics_app = make_metrics_app()
api.mount("/metrics", metrics_app)


def main():
    from .config import config

    logging.config.dictConfig(config["logging"])
    serve("epstats:api", settings.api, config["logging"])


if __name__ == "__main__":
    main()



import logging
import numpy as np
from fastapi import APIRouter, Depends, HTTPException
import asyncio
from concurrent.futures import ThreadPoolExecutor

from ..prometheus import get_prometheus_metric, Summary, Counter
from ..toolkit import Experiment as EvExperiment
from ..toolkit import Dao

from .req import Experiment
from .res import Result


_logger = logging.getLogger("epstats")
evaluation_duration_metric = get_prometheus_metric(
    "evaluation_duration_seconds", Summary, ["exp_id", "is_performance_test"]
)
query_duration_metric = get_prometheus_metric("query_duration_seconds", Summary, ["exp_id", "is_performance_test"])
stats_computation_duration_metric = get_prometheus_metric(
    "stats_computation_duration_seconds", Summary, ["exp_id", "is_performance_test"]
)
evaluation_errors_metric = get_prometheus_metric("evaluation_errors_total", Counter)
evaluation_successes_metric = get_prometheus_metric("evaluation_successes_total", Counter)
evaluation_requests_metric = get_prometheus_metric("evaluation_requests_total", Counter)


def get_evaluate_router(get_dao, get_executor_pool) -> APIRouter:
    def _evaluate(experiment: EvExperiment, dao: Dao):
        try:
            is_performance_test = experiment.query_parameters.get("is_performance_test") is True
            with evaluation_duration_metric.labels(experiment.id, is_performance_test).time():
                _logger.debug(f"Loading goals for experiment [{experiment.id}]")
                with query_duration_metric.labels(experiment.id, is_performance_test).time():
                    goals = dao.get_agg_goals(experiment).sort_values(["exp_variant_id", "goal"])
                    _logger.info(f"Retrieved {len(goals)} goals in experiment [{experiment.id}]")
                with stats_computation_duration_metric.labels(experiment.id, is_performance_test).time():
                    evaluation = experiment.evaluate_agg(goals)
                    evaluation_successes_metric.inc()
                _logger.info(
                    f"Evaluation response: [{experiment.id}]",
                    {
                        "metrics": (
                            evaluation.metrics.replace([np.inf, -np.inf], "inf")
                            .replace(np.nan, None)
                            .to_dict("records"),
                        )
                    },
                )
            return Result.from_evaluation(experiment, evaluation)
        except Exception as e:
            _logger.error(f"Cannot evaluate experiment [{experiment.id}] because of {e}")
            _logger.exception(e)
            evaluation_errors_metric.inc()
            raise HTTPException(
                status_code=500,
                detail=f"Cannot evaluate experiment [{experiment.id}] because of {e}",
            )

    router = APIRouter()

    @router.post("/evaluate", response_model=Result, tags=["Experiment Evaluation"])
    async def evaluate_experiment(
        experiment: Experiment,
        evaluation_pool: ThreadPoolExecutor = Depends(get_executor_pool),
        dao: Dao = Depends(get_dao),
    ):
        """
        Evaluates single `Experiment`.
        """
        _logger.info(f"Evaluation request: [{experiment.id}]", experiment.dict())
        evaluation_requests_metric.inc()
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(evaluation_pool, _evaluate, experiment.to_experiment(), dao)

    return router



from prometheus_client import Counter, Summary, REGISTRY
from typing import Union


application_prefix = "ep_stats_"


def get_prometheus_metric(metric_name: str, metric_type: type, labels: list[str] = None) -> Union[Counter, Summary]:
    if labels is None:
        labels = []
    metric_name_with_application_prefix = application_prefix + metric_name
    try:
        return metric_type(metric_name_with_application_prefix, "", labels)
    except ValueError:
        return REGISTRY._names_to_collectors[metric_name_with_application_prefix]
